// Reaktoro is a unified framework for modeling chemically reactive systems.
//
// Copyright © 2014-2022 Allan Leal
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library. If not, see <http://www.gnu.org/licenses/>.

#include "Database.hpp"

// C++ includes
#include <fstream>

// Reaktoro includes
#include <Reaktoro/Common/Algorithms.hpp>
#include <Reaktoro/Common/Exception.hpp>
#include <Reaktoro/Common/ParseUtils.hpp>
#include <Reaktoro/Core/Data.hpp>
#include <Reaktoro/Core/Support/DatabaseParser.hpp>

namespace Reaktoro {

struct Database::Impl
{
    /// The Species objects in the database.
    SpeciesList species;

    /// The Element objects in the database.
    ElementList elements;

    /// The additional data in the database whose type is known at runtime only.
    Any attached_data;

    /// The symbols of all elements already in the database.
    Set<String> element_symbols;

    /// The names of all species already in the database.
    Set<String> species_names;

    /// The species in the database grouped in terms of their aggregate state
    Map<AggregateState, SpeciesList> species_with_aggregate_state;

    /// Add an element in the database.
    auto addElement(Element const& element) -> void
    {
        if(element_symbols.find(element.symbol()) == element_symbols.end())
        {
            elements.append(element);
            element_symbols.insert(element.symbol());
        }
    }

    /// Add a species in the database.
    auto addSpecies(Species newspecies) -> void
    {
        // Ensure a unique name is used when storing this species!
        auto name = newspecies.name();
        auto unique_name = name;
        while(species_names.find(unique_name) != species_names.end())
            unique_name = unique_name + "!"; // keep adding symbol ! to the name such as H2O, H2O!, H2O!!, H2O!!! if many H2O are given

        // Replace name if not unique.
        if(name != unique_name) {
            newspecies = newspecies.withName(unique_name);
            warning(true, "Species should have unique names in Database, but species ", name, " "
                "violates this rule. The unique name ", unique_name, " has been assigned instead.");
        }

        // Replace aggregate state to Aqueous if undefined. If this default
        // aggregate state option is not appropriate for a given scenario,
        // ensure then that the correct aggregate state of the species has
        // already been set in the Species object using method
        // Species::withAggregateState. This permits species names such as H2O,
        // CO2, H2, O2 as well as ions H+, OH-, HCO3- to be considered as
        // aqueous species, without requiring explicit aggregate state suffix
        // aq as in H2O(aq), CO2(aq), H2(aq). For all other species, ensure an
        // aggregate state suffix is provided, such as H2O(l), H2O(g),
        // CaCO3(s), Fe3O4(s).
        if(newspecies.aggregateState() == AggregateState::Undefined)
            newspecies = newspecies.withAggregateState(AggregateState::Aqueous);

        // Append the new Species object in the species container
        species.append(newspecies);

        // Update the list of unique species names.
        species_names.insert(newspecies.name());

        // Update the container of elements with the Element objects in this new species.
        for(auto&& [element, coeff] : newspecies.elements())
            addElement(element);

        // Add the new species in the group of species with same aggregate state
        species_with_aggregate_state[newspecies.aggregateState()].push_back(newspecies);
    }

    /// Construct a reaction with given equation.
    auto reaction(String const& equation) const -> Reaction
    {
        return Reaction().withEquation(ReactionEquation(equation, species));
    }
};

Database::Database()
: pimpl(new Impl())
{}

Database::Database(Database const& other)
: pimpl(new Impl(*other.pimpl))
{}

Database::Database(Vec<Element> const& elements, Vec<Species> const& species)
: Database()
{
    for(auto const& x : elements)
        addElement(x);
    for(auto const& x : species)
        addSpecies(x);
}

Database::Database(Vec<Species> const& species)
: Database()
{
    for(auto const& x : species)
        addSpecies(x);
}

Database::~Database()
{}

auto Database::operator=(Database other) -> Database&
{
    pimpl = std::move(other.pimpl);
    return *this;
}

auto Database::clear() -> void
{
    *pimpl = Database::Impl();
}

auto Database::addElement(Element const& element) -> void
{
    pimpl->addElement(element);
}

auto Database::addSpecies(Species const& species) -> void
{
    pimpl->addSpecies(species);
}

auto Database::addSpecies(Vec<Species> const& species) -> void
{
    for(auto const& x : species)
        addSpecies(x);
}

auto Database::attachData(Any const& data) -> void
{
    pimpl->attached_data = data;
}

auto Database::extend(Database const& other) -> void
{
    for(auto const& element : other.elements())
        addElement(element);

    for(auto const& species : other.species())
        addSpecies(species);

    // TODO: Replace Any by Map<String, Any> so that it becomes easier/more intuitive to unify different attached data to Database objects.
    // pimpl->attached_data = ???;
}

auto Database::elements() const -> ElementList const&
{
    return pimpl->elements;
}

auto Database::species() const -> SpeciesList const&
{
    return pimpl->species;
}

auto Database::speciesWithAggregateState(AggregateState option) const -> SpeciesList
{
    auto it = pimpl->species_with_aggregate_state.find(option);
    if(it == pimpl->species_with_aggregate_state.end())
        return {};
    return it->second;
}

auto Database::element(String const& symbol) const -> Element const&
{
    return elements().getWithSymbol(symbol);
}

auto Database::species(String const& name) const -> Species const&
{
    return species().getWithName(name);
}

auto Database::reaction(String const& equation) const -> Reaction
{
    return pimpl->reaction(equation);
}

auto Database::attachedData() const -> Any const&
{
    return pimpl->attached_data;
}

auto Database::fromFile(String const& path) -> Database
{
    std::ifstream file(path);
    errorif(!file.is_open(),
        "Could not open file `", path, "`. Ensure the given file path "
        "is relative to the directory where your application is RUNNING "
        "(not necessarily where the executable is located!). Alternatively, "
        "try a full path to the file (e.g., "
        "in Windows, `C:\\User\\username\\mydata\\mydatabase.yaml`, "
        "in Linux, `/home/username/mydata/mydatabase.yaml`).");
    return fromStream(file);
}

auto Database::fromContents(String const& contents) -> Database
{
    auto doc = Data::parse(contents);
    DatabaseParser dbparser(doc);
    return dbparser;
}

auto Database::fromStream(std::istream& stream) -> Database
{
    auto doc = Data::parse(stream);
    DatabaseParser dbparser(doc);
    return dbparser;
}

} // namespace Reaktoro

